import asyncio
import functools
import collections
from routes import Mapper
from .webserver import WebServer
from .request import HttpRequest
from .response import HttpResponse

__all__ = ["Vase"]


_HANDLER_ATTR = '_vase__handler'
_BAG_ATTR = '_vase__bag'


class CallableDict(dict):
    "Dirty hack, so that routes does not stringify the dict"
    def __call__(self):
        pass


class Vase:
    def __init__(self, name):
        self._name = name
        self._routes = Mapper(register=False)
        self._endpoints = Mapper(register=False)

    @asyncio.coroutine
    def __call__(self, environ, start_response):
        match = self._routes.match(environ=environ)
        if match is None:
            return (yield from self._handle_404(environ, start_response))

        handler = match.pop(_HANDLER_ATTR)
        request = HttpRequest(environ)
        yield from request._maybe_init_post()

        data = yield from asyncio.coroutine(handler)(request)
        if isinstance(data, HttpResponse):
            response = data
        else:
            response = HttpResponse(data)
        
        return response(environ, start_response)

    def initialize_endpoint(self, environ):
        match = self._endpoints.match(environ=environ)
        if match is None:
            return None
        handler = match.pop(_HANDLER_ATTR)
        bag = match.pop(_BAG_ATTR)
        instance = handler()
        instance.bag = bag

        return instance

    @asyncio.coroutine
    def _handle_404(self, environ, start_response):
        data = 'Not found'.encode('utf-8')
        headers = (
            ('Content-Type', 'text/plain'),
            ('Content-Length', str(len(data)))
        )
        start_response(b'404 Not Found', headers)
        return [data]

    def route(self, *, path, methods=('get', 'post')):
        def wrap(func):
            self._routes.connect(path, **{_HANDLER_ATTR: func})
            return func

        return wrap

    def endpoint(self, *, path):
        def wrap(cls):
            self._endpoints.connect(path, **{_HANDLER_ATTR: cls, _BAG_ATTR: CallableDict()})
            return cls

        return wrap

    def run(self, *, host='0.0.0.0', port=3000, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        asyncio.async(loop.create_server(lambda: WebServer(loop=loop, app=self),
                    host, port))
        loop.run_forever()




STATUS_MAP = {
    100: 'CONTINUE',
    101: 'SWITCHING PROTOCOLS',
    102: 'PROCESSING',
    200: 'OK',
    201: 'CREATED',
    202: 'ACCEPTED',
    203: 'NON-AUTHORITATIVE INFORMATION',
    204: 'NO CONTENT',
    205: 'RESET CONTENT',
    206: 'PARTIAL CONTENT',
    207: 'MULTI-STATUS',
    208: 'ALREADY REPORTED',
    226: 'IM USED',
    300: 'MULTIPLE CHOICES',
    301: 'MOVED PERMANENTLY',
    302: 'FOUND',
    303: 'SEE OTHER',
    304: 'NOT MODIFIED',
    305: 'USE PROXY',
    306: 'RESERVED',
    307: 'TEMPORARY REDIRECT',
    400: 'BAD REQUEST',
    401: 'UNAUTHORIZED',
    402: 'PAYMENT REQUIRED',
    403: 'FORBIDDEN',
    404: 'NOT FOUND',
    405: 'METHOD NOT ALLOWED',
    406: 'NOT ACCEPTABLE',
    407: 'PROXY AUTHENTICATION REQUIRED',
    408: 'REQUEST TIMEOUT',
    409: 'CONFLICT',
    410: 'GONE',
    411: 'LENGTH REQUIRED',
    412: 'PRECONDITION FAILED',
    413: 'REQUEST ENTITY TOO LARGE',
    414: 'REQUEST-URI TOO LONG',
    415: 'UNSUPPORTED MEDIA TYPE',
    416: 'REQUESTED RANGE NOT SATISFIABLE',
    417: 'EXPECTATION FAILED',
    418: 'I\'M A TEAPOT',
    422: 'UNPROCESSABLE ENTITY',
    423: 'LOCKED',
    424: 'FAILED DEPENDENCY',
    426: 'UPGRADE REQUIRED',
    428: 'PRECONDITION REQUIRED',
    429: 'TOO MANY REQUESTS',
    431: 'REQUEST HEADER FIELDS TOO LARGE',
    500: 'INTERNAL SERVER ERROR',
    501: 'NOT IMPLEMENTED',
    502: 'BAD GATEWAY',
    503: 'SERVICE UNAVAILABLE',
    504: 'GATEWAY TIMEOUT',
    505: 'HTTP VERSION NOT SUPPORTED',
    506: 'VARIANT ALSO NEGOTIATES',
    507: 'INSUFFICIENT STORAGE',
    508: 'LOOP DETECTED',
    510: 'NOT EXTENDED',
    511: 'NETWORK AUTHENTICATION REQUIRED',
}

def status_line(status):
    line = STATUS_MAP[status]
    return "{} {}".format(status, line).encode('ascii')

class HttpResponse:
    def __init__(self, body, *, status=200, content_type='text/html'):
        if isinstance(body, str):
            body = body.encode('utf-8')
        self._body = body
        self._status = int(status)
        self._headers = [
            ('Content-Encoding', 'UTF-8'),
            ('Content-Type', content_type),
            ('Content-Length', str(len(body))),
        ]
        self._content_type = content_type

    def __call__(self, environ, start_response):
        status = status_line(self._status)
        start_response(status, self._get_headers())
        return [self._body]


    def _get_headers(self):
        return [(x.encode('ascii'), y.encode('ascii')) for x, y in self._headers]




import asyncio
from asyncio.streams import StreamWriter
from urllib.parse import unquote
import sys

from .stream import LimitedReader
from .exceptions import BadRequestException

_DEFAULT_EXHAUST = 2**16
DELIMITER = b'\r\n'


class HttpWriter(StreamWriter):
    delimiter = b'\r\n'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._headers_sent = False
        self._status_written = False
    
    def write_status(self, status, version=b'1.1'):
        assert not self._status_written, "Status has already been written"
        self._status_written = True
        self.write(b'HTTP/' + version + b' ' + status + self.delimiter)

    def write_header(self, name, value):
        assert not self._headers_sent, "Headers have already been sent"
        if type(name) is not bytes:
            name = name.encode('ascii')
        if type(value) is not bytes:
            value = value.encode('ascii')
        self.write(name + b': ' + value + self.delimiter)

    def write_headers(self, headers):
        for name, value in headers:
            self.write_header(name, value)

    def _maybe_finalize_headers(self):
        if not self._headers_sent:
            self.write(self.delimiter)
        self._headers_sent = True

    def write_body(self, data):
        self._maybe_finalize_headers()
        self.write(data)

    def writelines(self, data):
        self._maybe_finalize_headers()
        super().writelines(data)

    def restore(self):
        self._headers_sent = False
        self._status_written = False

    @property
    def status_written(self):
        return self._status_written
    

class WsgiParser:
    @staticmethod
    @asyncio.coroutine
    def parse(reader):
        l = yield from reader.readline()
        if not l:
            return
        l = l.rstrip(DELIMITER)
        try:
            method, url, version = (x.decode('ascii') for x in l.split(b' '))
            if version not in ('HTTP/1.1', 'HTTP/1.0'):
                raise ValueError('Unsupported http version {}'.format(version))
        except ValueError:
            raise BadRequestException()

        try:
            path, query = url.split('?')
        except ValueError:
            path = url
            query = ''

        peer = reader._transport.get_extra_info('peername')
        sslctx = reader._transport.get_extra_info('sslcontext')
        ip, port = peer[0], peer[1]

        headers = {}
        while True:
            l = yield from reader.readline()
            if l == DELIMITER:
                break

            l = l.rstrip(DELIMITER)
            try:
                name, value = (x.strip() for x in l.split(b':', 1))
            except ValueError:
                raise BadRequestException()
            else:
                hname = 'HTTP_{}'.format(name.decode('ascii').replace('-', '_').upper())
                if hname not in headers:
                    headers[hname] = value.decode('ascii')
                else:
                    sep = ','
                    if hname == 'HTTP_COOKIE':
                        sep = ';'
                    headers[hname] += "{}{}".format(sep, value)
        try:
            content_length = int(headers.get('HTTP_CONTENT_LENGTH', '0'))
        except ValueError:
            content_length = 0
        environ = {
            'REQUEST_METHOD': method,
            'SCRIPT_NAME': '',
            'PATH_INFO': unquote(path),
            'QUERY_STRING': unquote(query),
            'REMOTE_ADDR': ip,
            'SERVER_PROTOCOL': version,
            'REMOTE_PORT': port,
            'wsgi.input': LimitedReader(reader, content_length),
            'wsgi.error': sys.stderr,
            'wsgi.version': (1, 0),
            'wsgi.multithread': False,
            'wsgi.multiprocess': False,
            'wsgi.run_once': False,
            'wsgi.url_scheme': 'http' if sslctx is None else 'https',
        }
        environ.update(headers)
        return environ





import asyncio
from .exceptions import BadRequestException
from .http import (
    HttpWriter,
    WsgiParser,
)
from hashlib import sha1
from base64 import b64encode
from .websocket import (
    WebSocketWriter,
    WebSocketParser,
    MAGIC,
    OpCode,
)
from .log import logger
import sys

def should_close_conn_immediately(environ):
    result = False
    if environ['SERVER_PROTOCOL'].lower() == 'http/1.0':
        result = True
    conn_header = environ.get('HTTP_CONNECTION', '').lower()
    if conn_header == 'keep-alive':
        result = False
    elif conn_header == 'close':
        result = True
    return result


WEBSOCKET_HEADERS = (
    'HTTP_UPGRADE',
    'HTTP_CONNECTION',
    'HTTP_SEC_WEBSOCKET_KEY',
    'HTTP_SEC_WEBSOCKET_VERSION',
)

def is_websocket_request(environ):
    upgrade = environ.get('HTTP_UPGRADE', '').lower()
    connection = environ.get('HTTP_CONNECTION', '').lower()
    key = environ.get('HTTP_SEC_WEBSOCKET_KEY', None)
    version = environ.get('HTTP_SEC_WEBSOCKET_VERSION', None)
    if upgrade != 'websocket' or 'upgrade' not in connection or key is None or version != '13':
        return False
    return True


def _maybe_log_exception(task):
    try:
        task.result()
    except:
        logger.exception("An exception ocurred while serving request")

class WebServer(asyncio.StreamReaderProtocol):
    KEEP_ALIVE = 20
    def __init__(self, *args, app, loop=None):
        self._app = app
        stream_reader = asyncio.StreamReader(loop=loop)
        super().__init__(stream_reader, None, loop)
        self.reader = self._stream_reader
        self.h_timeout = None
        self._timeout_disabled = False
        self._in_ws_mode = False

    def connection_made(self, transport):
        self._transport = transport
        self.reader.set_transport(transport)

        self.writer = HttpWriter(transport, self,
                                self._stream_reader,
                                self._loop)

        task = asyncio.async(self._handle_client())
        task.add_done_callback(_maybe_log_exception)

        self._reset_timeout()

    def connection_lost(self, exc):
        self.writer = None
        self._stop_timeout()
        super().connection_lost(exc)

    def data_received(self, data):
        self._reset_timeout()
        super().data_received(data)

    @asyncio.coroutine
    def _handle_client(self):
        while True:
            try:
                env = yield from WsgiParser.parse(self.reader)
            except BadRequestException as e:
                self._bad_request(self.writer, e)
                self.writer.close()
                break
            # connection has been closed
            if env is None:
                break

            try:
                yield from self.handle_request(env)
            except BadRequestException as e:
                self._bad_request(self.writer, e)
            except Exception as e:
                self._transport.close()
                raise
            if should_close_conn_immediately(env):
                self.writer.close()
            else:
                yield from env['wsgi.input'].read()
                self.writer.restore()

    def _reset_timeout(self):
        self._stop_timeout()
        if not self._timeout_disabled:
            self.h_timeout = self._loop.call_later(
                self.KEEP_ALIVE, self._handle_timeout)

    def _stop_timeout(self):
        if self.h_timeout is not None:
            self.h_timeout.cancel()
            self.h_timeout = None

    def _handle_timeout(self):
        if not self._in_ws_mode:
            self.writer.close()
        else:
            self._ws_handler.transport._write_ws(OpCode.ping, b'')
            self._reset_timeout()

    @asyncio.coroutine
    def handle_request(self, environ):
        if is_websocket_request(environ):
            yield from self._handle_websocket(environ)
            return
        yield from self._handle_wsgi(environ)

    @asyncio.coroutine
    def _handle_wsgi(self, environ):
        def start_response(status, headers, exc_info=None):
            self.writer.write_status(status)
            self.writer.write_headers(headers)

            return self.writer

        result = yield from self._app(environ, start_response)
        assert self.writer.status_written, 'Wsgi app did not call start_response!'
        if not should_close_conn_immediately(environ):
            self.writer.write_header('Connection', 'Keep-Alive')
        self.writer.writelines(result)

    @asyncio.coroutine
    def _handle_websocket(self, environ):
        writer = self.writer
        handler = self._app.initialize_endpoint(environ)
        if handler is None:
            data = "Not found".encode('utf-8')
            writer.write_status(b'404 Not Found')
            writer.write_headers((
                ('Content-Length', str(len(data))),
            ))
            writer.write_body(data)
            return

        handler.transport = WebSocketWriter(self._transport)
        handler.loop = self._loop

        if not (yield from asyncio.coroutine(handler.authorize_request)(environ)):
            writer.write_status(b'401 Anauthorized')
            writer.write_body(b'')
            return            

        key = environ['HTTP_SEC_WEBSOCKET_KEY']

        accept = sha1(key.encode('ascii') + MAGIC).digest()
        writer.write_status(b'101 Switching Protocols')
        writer.write_headers((
            ('Upgrade', 'websocket',),
            ('Connection', 'Upgrade'),
            ('Sec-WebSocket-Accept', b64encode(accept))
        ))
        writer.write_body(b'')
        yield from self._switch_protocol(handler)

    def _switch_protocol(self, handler):
        #self._disable_timeout()
        self._ws_handler = handler
        self._ws_handler.on_connect()
        self._in_ws_mode = True

        #self._stream_reader = StreamReader(loop=self._loop)
        #self._stream_reader.set_transport(self._transport)

        yield from self._parse_messages()

    @asyncio.coroutine
    def _parse_messages(self):
        parser = WebSocketParser(self._stream_reader)
        while True:
            msg = yield from parser.get_message()
            if msg is None:
                return
            if msg.is_ctrl:
                if msg.opcode == OpCode.close:
                    self._ws_handler.transport._write_ws(OpCode.close, b'')
                    self._transport.close()
                    return
                elif msg.opcode == OpCode.ping:
                    self._ws_handler.transport._write_ws(OpCode.pong, b'')
            else:
                self._ws_handler.on_message(msg.payload)

    def _bad_request(self, writer, exc):
        writer.write_status(b'400 Bad Request')
        writer.write_header('Content-Type', 'text/plain')
        body = b''
        if exc is not None and len(exc.args):
            body = str(exc.args[0]).encode('utf-8')
        writer.write_header('Content-Length', str(len(body)))
        writer.write_body(body)


if __name__ == '__main__':
    loop = asyncio.get_event_loop()

    asyncio.async(loop.create_server(lambda: WebServer(loop=loop), 'localhost', 3000))
    loop.run_forever()



import sys
from setuptools import setup

install_requires = [
    'routes==2.0'
]
if sys.version_info < (3, 4):
    install_requires.append('enum34')
    install_requires.append('asyncio==0.2.1')

setup(
    name="Vase",
    version="0.1.3",
    author = "Vladimir Kryachko",
    author_email = "v.kryachko@gmail.com",
    description = "Async Web framework based on Tulip/asyncio",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
    ],
    install_requires = install_requires,
    packages=["vase"],
)



import asyncio
import urllib.parse
from http.cookies import CookieError, SimpleCookie

_FORM_URLENCODED = 'application/x-www-form-urlencoded'


class MultiDict(dict):
    def __getitem__(self, key):
        value = super().__getitem__(key)
        if type(value) is list and value:
            return value[0]
        raise KeyError(key)

    def __setitem__(self, key, value):
        super().__setitem__(key, [value])

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def getlist(self, key, default=[]):
        return super().get(key, default)

    def items(self):
        for key in self:
            yield key, self[key]

    def values(self):
        for key in self:
            yield self[key]

    def lists(self):
        return super().items()

    def __repr__(self):
        return "<MultiDict {}>".format(super().__repr__())


class HttpRequest:
    def __init__(self, environ):
        self.ENV = environ
        self.method = environ["REQUEST_METHOD"].upper()
        self.path = environ['PATH_INFO']
        self.path_info = environ['PATH_INFO']

        self.POST = MultiDict()
        self._cookies = None
        self._post_inited = False

    @property
    def GET(self):
        if self._get is None:
            self._get = MultiDict(urllib.parse.parse_qs(self.ENV.get('QUERY_STRING')))
        return self._get


    @property
    def COOKIES(self):
        if self._cookies is None:
            try:
                c = SimpleCookie(self.ENV.get('HTTP_COOKIE', ''))
            except CookieError:
                self._cookies = {}
            else:
                res = {}
                for name, value in c.items():
                    res[name] = value.value
                self._cookies = res
        return self._cookies

    def _has_form(self):
        return self.ENV.get('HTTP_CONTENT_TYPE', '') == _FORM_URLENCODED

    @asyncio.coroutine
    def _maybe_init_post(self):
        if self._post_inited:
            return
        if self._has_form():
            body = yield from self.ENV['wsgi.input'].read()
            self.POST = MultiDict(urllib.parse.parse_qs(body.decode('utf-8')))




from asyncio import coroutine


class LimitedReader:
    def __init__(self, reader, limit):
        self._reader = reader
        self._limit = limit
        self._read_count = 0

    @coroutine
    def read(self, n=-1):
        if not n or not self.bytes_left:
            return b''
        if n < 0 or n > self.bytes_left:
            n = self.bytes_left
        self._read_count += n
        return (yield from self._reader.read(n))

    @property
    def bytes_left(self):
        left = self._limit - self._read_count
        if left < 0:
            left = 0
        return left



import asyncio
from asyncio.streams import StreamWriter, StreamReader
import ssl
from .exceptions import BadRequestException
from base64 import b64encode
from hashlib import sha1
import collections
import struct
from enum import Enum

MAGIC = b'258EAFA5-E914-47DA-95CA-C5AB0DC85B11'

class OpCode(Enum):
    """
    WebSocket opcodes as defined in RFC 6455
    """
    continuation = 0
    text = 1
    binary = 2
    close = 8
    ping = 9
    pong = 10

    @property
    def is_ctrl(self):
        return self.value > 0x7


class Frame:
    """
    WebSocket frame
    """
    __slots__ = ('fin', 'opcode', 'payload')
    def __init__(self, fin, opcode, payload):
        self.fin = fin
        self.opcode = opcode
        self.payload = payload

    @property
    def is_ctrl(self):
        """
        Returns True if it is a control frame
        """
        return self.opcode.is_ctrl


class Message:
    __slots__ = ('opcode', 'payload', 'ext_data')
    def __init__(self, opcode, payload, ext_data):
        self.opcode = opcode
        self.payload = payload
        self.ext_data = ext_data

    @property
    def is_ctrl(self):
        """
        Returns True if it is a control frame
        """
        return self.opcode.is_ctrl

    @classmethod
    def close_message(cls, reason, *, message=b''):
        """
        Creates a 'close' message with specified reason code and optional message
        """
        return cls(Opcode.close, b'', b'')



class WebSocketFormatException(Exception):
    def __init__(self, *args):
        if len(args) > 0:
            self.reason = args[0]
        else:
            self.reasons = None
        super().__init__(*args)


class WebSocketParser:
    """
    This object is instantiated for each connection
    """
    def __init__(self, reader):
        self._reader = reader
        self._frames = collections.deque()

    @asyncio.coroutine
    def get_message(self):
        while True:
            frame = yield from self.parse_frame(self._reader)
            if frame is None:
                return
            if frame.is_ctrl:
                return Message(frame.opcode, frame.payload, '')

            if not self._frames and frame.opcode not in (OpCode.binary, OpCode.text):
                raise WebSocketFormatException("The first data frame must be either 'binary' or 'text'")

            if self._frames and frame.opcode != OpCode.continuation:
                raise WebSocketFormatException("Frames belonging to different messages cannot be interleaved")
            self._frames.append(frame)
            if frame.fin:
                return self._build_message()

    def _build_message(self):
        buf = []
        frame = self._frames.popleft()
        opcode = frame.opcode
        buf.append(frame.payload)
        while self._frames:
            frame = self._frames.popleft()
            buf.append(frame.payload)
        payload = b''.join(buf)
        if opcode == OpCode.text:
            payload = payload.decode('utf-8')
        return Message(opcode, payload, b'')

    @classmethod
    def parse_frame(cls, reader):
        data = yield from reader.read(2)
        if not data:
            return None
        first_byte, second_byte = struct.unpack('!BB', data)

        fin = (first_byte >> 7) & 1
        rsv1 = (first_byte >> 6) & 1
        rsv2 = (first_byte >> 5) & 1
        rsv3 = (first_byte >> 4) & 1
        opcode = first_byte & 0xf

        length = (second_byte) & 0x7f

        try:
            opcode = OpCode(opcode)
        except ValueError:
            raise WebSocketFormatException("Unknown opcode received '0x{:X}'".format(opcode))

        if rsv1 or rsv2 or rsv3:
            raise WebSocketFormatException("Reserved bits must be set to 0")

        if opcode.is_ctrl:
            if not fin:
                raise WebSocketFormatException("Control frames MUST NOT be fragmented")
            if length > 125:
                raise WebSocketFormatException("All control frames MUST have a payload length of 125 bytes or less")

        has_mask = (second_byte >> 7) & 1

        if not has_mask:
            raise WebSocketFormatException("Clients MUST mask their frames")


        if length == 126:
            data = yield from reader.read(2)
            if not data:
                return None
            length = struct.unpack_from('!H', data)[0]
        elif length == 127:
            data = yield from reader.read(8)
            if not data:
                return None
            length = struct.unpack_from('!Q', data)[0]

        mask = yield from reader.read(4)
        if not mask:
            return None

        if length:
            payload = yield from reader.read(length)
            if not payload:
                return None
        else:
            payload = b''

        payload = bytes(b ^ mask[i % 4] for i, b in enumerate(payload))

        return Frame(fin, opcode, payload)


class WebSocketWriter(StreamWriter):
    def __init__(self, transport, *, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self._loop = loop
        self._transport = transport

    def send(self, msg, later=False):
        opcode = OpCode.text
        if isinstance(msg, bytes):
            opcode = OpCode.binary
        else:
            msg = msg.encode('utf-8')

        self._write_ws(opcode, msg, later)

    def _write_ws(self, opcode, payload, later=False):
        first_byte = bytes([(1<<7) | opcode.value])
        length = len(payload)
        if length <= 125:
            len_bytes = bytes([length])
        elif length < 2**16:
            len_bytes = struct.pack('!BH', 126, length)
        else:
            len_bytes = struct.pack('!BQ', 127, length)
        msg = b''.join([first_byte, len_bytes, payload])
        
        if not later:
            self._transport.write(msg)
        else:
            self._loop.call_soon(lambda: self._transport.write(msg))
